@@ -25,8 +25,8 @@
 #define HAVE_ADDR_MASK
 #define HAVE_ADDR_TYPE
 #define HAVE_ALTQ
-#define HAVE_RULE_ANCHOR
 #define HAVE_RULE_TOS
+#define HAVE_OP_RRG
 #endif
 
 #if OS_LEVEL > 33
@@ -35,4 +35,13 @@
 #define HAVE_RULE_NATPASS
 #endif
 
+#if OS_LEVEL > 35
+#define HAVE_NEG
+#define HAVE_RULESETS
+#endif
+
+#if OS_LEVEL > 37
+#define HAVE_INOUT_COUNT_RULES
+#endif
+
 #endif
