@@ -1,6 +1,6 @@
-/* $Id: pftop.c,v 1.77.2.5 2003/08/04 21:34:10 canacar Exp $	 */
+/* $Id: pftop.c,v 1.77.4.1 2006/01/11 01:59:17 canacar Exp $	 */
 /*
- * Copyright (c) 2001 Can Erkin Acar
+ * Copyright (c) 2006 Can Erkin Acar
  * Copyright (c) 2001 Daniel Hartmeier
  * All rights reserved.
  *
@@ -149,7 +149,7 @@ void sort_states(void);
 void print_states(void);
 
 int select_rules(void);
-int read_rules(int dev);
+int read_rules(int);
 void print_rules(void);
 
 #ifdef HAVE_ALTQ
@@ -257,11 +257,13 @@ typedef struct {
 #define FLD_PKTSPS  34
 #define FLD_BYTESPS 35
 
-
 #ifndef HAVE_RULE_LABELS
 #define PF_RULE_LABEL_SIZE 20
 #endif
 
+#define MIN_LABEL_SIZE 5
+#define ANCHOR_FLD_SIZE 12
+
 field_def fields[] = {
 	{"SRC", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0},
 	{"DEST", 20, 45, 1, FLD_ALIGN_LEFT, -1, 0, 0},
@@ -269,12 +271,12 @@ field_def fields[] = {
 	{"STATE", 5, 23, 18, FLD_ALIGN_COLUMN, -1, 0, 0},
 	{"AGE", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"EXP", 5, 9, 4, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"PR ", 4, 4, 1, FLD_ALIGN_LEFT, -1, 0, 0},
+	{"PR ", 4, 9, 1, FLD_ALIGN_LEFT, -1, 0, 0},
 	{"DIR", 1, 3, 2, FLD_ALIGN_CENTER, -1, 0, 0},
 	{"PKTS", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"BYTES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"RULE", 2, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"LABEL", 20, PF_RULE_LABEL_SIZE, 1, FLD_ALIGN_LEFT, -1, 0, 0},
+	{"LABEL", MIN_LABEL_SIZE, MIN_LABEL_SIZE, 1, FLD_ALIGN_LEFT, -1, 0, 0},
 	{"STATES", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"EVAL", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"ACTION", 1, 8, 4, FLD_ALIGN_LEFT, -1, 0, 0},
@@ -289,14 +291,14 @@ field_def fields[] = {
 	{"PEAK", 5, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"ANCHOR", 6, 16, 1, FLD_ALIGN_LEFT, -1, 0, FLD_FLAG_HIDDEN},
 	{"QUEUE", 15, 30, 1, FLD_ALIGN_LEFT, -1, 0, 0},
-	{"BANDW", 5, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
+	{"BW", 4, 5, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"SCH", 3, 4, 1, FLD_ALIGN_LEFT, -1, 0, 0},
 	{"PRIO", 1, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"DROP_P", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"DROP_B", 6, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"QLEN", 4, 4, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"BORROW", 4, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
-	{"SUSPENDS", 4, 8, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
+	{"BORROW", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
+	{"SUSPENDS", 4, 6, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"P/S", 3, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0},
 	{"B/S", 4, 7, 1, FLD_ALIGN_RIGHT, -1, 0, 0}
 };
@@ -310,30 +312,26 @@ int view0[] = {
 
 int view1[] = {
 	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_GW, FLD_STATE, FLD_AGE,
-	FLD_EXP, FLD_PKTS, FLD_BYTES, FLD_SI, FLD_SP, FLD_SA, FLD_RULE,
-	FLD_ANCHOR, -1
+	FLD_EXP, FLD_PKTS, FLD_BYTES, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, -1
 };
 
 int view2[] = {
 	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_STATE, FLD_AGE, FLD_EXP,
-	FLD_PKTS, FLD_BYTES, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_ANCHOR,
-	FLD_GW, -1
+	FLD_PKTS, FLD_BYTES, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_GW, -1
 };
 
 int view3[] = {
 	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_AGE, FLD_EXP, FLD_PKTS,
-	FLD_BYTES, FLD_STATE, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_ANCHOR,
-	FLD_GW, -1
+	FLD_BYTES, FLD_STATE, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_GW, -1
 };
 
 int view4[] = {
 	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_PKTS, FLD_BYTES, FLD_STATE,
-	FLD_AGE, FLD_EXP, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_ANCHOR,
-	FLD_GW, -1
+	FLD_AGE, FLD_EXP, FLD_SI, FLD_SP, FLD_SA, FLD_RULE, FLD_GW, -1
 };
 
 int view5[] = {
-	FLD_RULE, FLD_ACTION, FLD_DIR, FLD_LOG, FLD_QUICK, FLD_IF,
+	FLD_RULE, FLD_ANCHOR, FLD_ACTION, FLD_DIR, FLD_LOG, FLD_QUICK, FLD_IF,
 	FLD_PROTO, FLD_KST, FLD_PKTS, FLD_BYTES, FLD_STATS, FLD_STMAX,
 	FLD_RINFO, -1
 };
@@ -341,13 +339,12 @@ int view5[] = {
 int view6[] = {
 	FLD_RULE, FLD_LABEL, FLD_PKTS, FLD_BYTES, FLD_STATS, FLD_STMAX,
 	FLD_ACTION, FLD_DIR, FLD_LOG, FLD_QUICK, FLD_IF, FLD_PROTO,
-	FLD_KST, -1
+	FLD_ANCHOR, FLD_KST, -1
 };
 
 int view7[] = {
 	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST,  FLD_SI, FLD_SP, FLD_SA,
-	FLD_BYTES, FLD_STATE, FLD_PKTS, FLD_AGE, FLD_EXP, FLD_RULE,
-	FLD_ANCHOR, FLD_GW, -1
+	FLD_BYTES, FLD_STATE, FLD_PKTS, FLD_AGE, FLD_EXP, FLD_RULE, FLD_GW, -1
 };
 
 int view8[] = {
@@ -699,7 +696,8 @@ field_setup(void)
 	for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
 		n = *v;
 
-		if (fields[n].flags & FLD_FLAG_HIDDEN)
+		if (fields[n].flags & FLD_FLAG_HIDDEN ||
+		    fields[n].max_width == 0)
 			continue;
 		if (width <= 1)
 			break;
@@ -721,7 +719,8 @@ field_setup(void)
 		change = 0;
 		for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
 			n = *v;
-			if (fields[n].flags & FLD_FLAG_HIDDEN)
+			if (fields[n].flags & FLD_FLAG_HIDDEN ||
+			    fields[n].max_width == 0)
 				continue;
 			if ((fields[n].width < fields[n].max_width) &&
 			    (fields[n].increment <= width)) {
@@ -740,7 +739,8 @@ field_setup(void)
 	st = 0;
 	for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
 		n = *v;
-		if (fields[n].flags & FLD_FLAG_HIDDEN)
+		if (fields[n].flags & FLD_FLAG_HIDDEN ||
+		    fields[n].max_width == 0)
 			continue;
 		if (fields[n].start < 0) break;
 		fields[n].start = st;
@@ -864,17 +864,17 @@ int
 sort_size_callback(const void *s1, const void *s2)
 {
 #ifdef HAVE_INOUT_COUNT
-	u_int32_t b1 = state_buf[* (u_int32_t *) s1].bytes[0] + 
+	u_int64_t b1 = state_buf[* (u_int32_t *) s1].bytes[0] + 
 		state_buf[* (u_int32_t *) s1].bytes[1];
-	u_int32_t b2 = state_buf[* (u_int32_t *) s2].bytes[0] + 
+	u_int64_t b2 = state_buf[* (u_int32_t *) s2].bytes[0] + 
 		state_buf[* (u_int32_t *) s2].bytes[1];
 #else
-	u_int32_t b1 = state_buf[* (u_int32_t *) s1].bytes;
-	u_int32_t b2 = state_buf[* (u_int32_t *) s2].bytes;
+	u_int64_t b1 = state_buf[* (u_int32_t *) s1].bytes;
+	u_int64_t b2 = state_buf[* (u_int32_t *) s2].bytes;
 #endif
 	if (b2 > b1)
 		return sortdir;
-	if (b1 < b2)
+	if (b2 < b1)
 		return -sortdir;
 	return 0;
 }
@@ -883,17 +883,17 @@ int
 sort_pkt_callback(const void *s1, const void *s2)
 {
 #ifdef HAVE_INOUT_COUNT
-	u_int32_t p1 = state_buf[* (u_int32_t *) s1].packets[0] + 
+	u_int64_t p1 = state_buf[* (u_int32_t *) s1].packets[0] + 
 		state_buf[* (u_int32_t *) s1].packets[1];
-	u_int32_t p2 = state_buf[* (u_int32_t *) s2].packets[0] + 
+	u_int64_t p2 = state_buf[* (u_int32_t *) s2].packets[0] + 
 		state_buf[* (u_int32_t *) s2].packets[1];
 #else
-	u_int32_t p1 = state_buf[* (u_int32_t *) s1].packets;
-	u_int32_t p2 = state_buf[* (u_int32_t *) s2].packets;
+	u_int64_t p1 = state_buf[* (u_int32_t *) s1].packets;
+	u_int64_t p2 = state_buf[* (u_int32_t *) s2].packets;
 #endif
 	if (p2 > p1)
 		return sortdir;
-	if (p1 < p2)
+	if (p2 < p1)
 		return -sortdir;
 	return 0;
 }
@@ -1309,7 +1309,6 @@ print_header(struct pf_status *status)
 
 	time(&t);
 	tp = localtime(&t);
-
 	if (tp) {
 		int len;
 
@@ -1474,7 +1473,7 @@ print_fld_state(int fid, unsigned int proto, unsigned int s1, unsigned int s2)
 }
 
 void
-print_fld_size(int fid, unsigned int size)
+print_fld_sdiv(int fid, u_int64_t size, int div)
 {
 	int len;
 
@@ -1485,26 +1484,26 @@ print_fld_size(int fid, unsigned int size)
 		return;
 
 	tb_start();
-	if (tbprintf("%u", size) <= len)
+	if (tbprintf("%llu", size) <= len)
 		goto ok;
 
-	size /= 1024;
+	size /= div;
 	tb_start();
-	if (tbprintf("%uK", size) <= len)
+	if (tbprintf("%lluK", size) <= len)
 		goto ok;
 	if (size == 0)
 		goto err;
 
-	size /= 1024;
+	size /= div;
 	tb_start();
-	if (tbprintf("%uM", size) <= len)
+	if (tbprintf("%lluM", size) <= len)
 		goto ok;
 	if (size == 0)
 		goto err;
 
-	size /= 1024;
+	size /= div;
 	tb_start();
-	if (tbprintf("%uG", size) <= len)
+	if (tbprintf("%lluG", size) <= len)
 		goto ok;
 	
 err:
@@ -1517,6 +1516,12 @@ ok:
 	tb_end();
 }
 
+__inline void
+print_fld_size(int fid, u_int64_t size)
+{
+	print_fld_sdiv(fid, size, 1024);
+}
+
 void
 print_fld_rate(int fid, double rate)
 {
@@ -1527,6 +1532,16 @@ print_fld_rate(int fid, double rate)
 	}
 }
 
+void
+print_fld_bw(int fid, double bw)
+{
+	if (bw < 0) {
+		print_fld_str(fid, "*");
+	} else {
+		print_fld_sdiv(fid, bw, 1000);
+	}
+}
+
 void
 print_fld_uint(int fid, unsigned int size)
 {
@@ -1591,7 +1606,7 @@ print_state(struct pf_state * s, struct sc_ent * ent)
 	print_fld_age(FLD_EXP, s->expire);
 #ifdef HAVE_INOUT_COUNT
 	{
-		u_int32_t sz = s->bytes[0]+s->bytes[1];
+		u_int64_t sz = s->bytes[0]+s->bytes[1];
 
 		print_fld_size(FLD_PKTS, s->packets[0]+s->packets[1]);
 		print_fld_size(FLD_BYTES, sz);
@@ -1645,7 +1660,7 @@ disp_update(int dev)
 		dispstart = 0;
 
 	if (ioctl(dev, DIOCGETSTATUS, &status)) {
-		warnx("DIOCGETSTATUS");
+		warn("DIOCGETSTATUS");
 		return (-1);
 	}
 
@@ -1673,21 +1688,14 @@ select_rules(void)
 	return (0);
 }
 
-int
-read_rules(int dev)
+
+void
+add_rule_alloc(u_int32_t nr)
 {
-	struct pfioc_rule pr;
-	u_int32_t nr;
+	if (nr == 0)
+		return;
 
-	memset(&pr, 0, sizeof(pr));
-	if (ioctl(dev, DIOCGETRULES, &pr)) {
-		warnx("DIOCGETRULES");
-		return (-1);
-	}
-
-	num_disp = num_rules = pr.nr;
-
-	if (num_rules == 0) return (0);
+	num_rules += nr;
 
 	if (rules == NULL) {
 		rules = malloc(num_rules * sizeof(struct pf_rule));
@@ -1700,19 +1708,232 @@ read_rules(int dev)
 			err(1, "realloc");
 		alloc_rules = num_rules;
 	}
+}
 
-	for (nr = 0; nr < num_rules; ++nr) {
+#ifdef HAVE_RULE_LABELS
+int label_length;
+#endif
+
+int
+read_anchor_rules(int dev, char *anchor)
+{
+	struct pfioc_rule pr;
+	u_int32_t nr, num, off;
+
+	memset(&pr, 0, sizeof(pr));
+#ifdef HAVE_RULESETS
+	strlcpy(pr.anchor, anchor, sizeof(pr.anchor));
+#endif
+	if (ioctl(dev, DIOCGETRULES, &pr)) {
+		warn(anchor);
+		return (-1);
+	}
+
+	off = num_rules;
+	num = pr.nr;
+	add_rule_alloc(num);
+
+	for (nr = 0; nr < num; ++nr) {
 		pr.nr = nr;
 		if (ioctl(dev, DIOCGETRULE, &pr)) {
-			warnx("DIOCGETRULE");
+			warn("DIOCGETRULE");
 			return (-1);
 		}
-		rules[nr] = pr.rule;
+#ifdef HAVE_RULESETS
+		/* XXX overload pr.anchor, to store a pointer to
+		 * anchor name */
+		pr.rule.anchor = (struct pf_anchor *) anchor;
+#endif
+#ifdef HAVE_RULE_LABELS
+		{
+			int len = strlen(pr.rule.label);
+			if (len > label_length)
+				label_length = len;
+		}
+#endif
+		rules[off + nr] = pr.rule;
+	}
+
+	return (num);
+}
+
+#ifdef HAVE_RULESETS
+struct anchor_name {
+	char name[MAXPATHLEN];
+	struct anchor_name *next;
+	u_int32_t ref;
+};
+
+struct anchor_name *anchor_root = NULL;
+struct anchor_name *anchor_end = NULL;
+struct anchor_name *anchor_free = NULL;
+
+struct anchor_name*
+alloc_anchor_name(const char *path)
+{
+	struct anchor_name *a;
+
+	a = anchor_free;
+	if (a == NULL) {
+		a = (struct anchor_name *)malloc(sizeof(struct anchor_name));
+		if (a == NULL)
+			return (NULL);
+	} else
+		anchor_free = a->next;
+
+	if (anchor_root == NULL)
+		anchor_end = a;
+
+	a->next = anchor_root;
+	anchor_root = a;
+
+	a->ref = 0;
+	strlcpy(a->name, path, sizeof(a->name));
+	return (a);
+}
+
+void
+reset_anchor_names(void)
+{
+	if (anchor_end == NULL)
+		return;
+
+	anchor_end->next = anchor_free;
+	anchor_free = anchor_root;
+	anchor_root = anchor_end = NULL;
+}
+
+struct pfioc_ruleset ruleset;
+char *rs_end = NULL;
+
+int
+read_rulesets(int dev, const char *path)
+{
+	char *pre;
+	struct anchor_name *a;
+	u_int32_t nr, ns;
+	int len;
+
+	if (path == NULL)
+		ruleset.path[0] = '\0';
+	else if (strlcpy(ruleset.path, path, sizeof(ruleset.path)) >= 
+	    sizeof(ruleset.path))
+		 return (-1);
+
+	/* a persistent storage for anchor names */
+	a = alloc_anchor_name(ruleset.path);
+	if (a == NULL)
+		return (-1);
+
+	len = read_anchor_rules(dev, a->name);
+	if (len < 0)
+		return (-1);
+
+	a->ref += len;
+
+	if (ioctl(dev, DIOCGETRULESETS, &ruleset)) {
+		warn("DIOCGETRULESETS");
+		return (-1);
+	}
+
+	ns = ruleset.nr;
+
+	if (rs_end == NULL)
+		rs_end = ruleset.path + sizeof(ruleset.path);
+
+	/* 'pre' tracks the previous level on the anchor */
+	pre = strchr(ruleset.path, 0);
+	len = rs_end - pre;
+	if (len < 1)
+		return (-1);
+	--len;
+
+	for (nr = 0; nr < ns; ++nr) {
+		ruleset.nr = nr;
+		if (ioctl(dev, DIOCGETRULESET, &ruleset)) {
+			warn("DIOCGETRULESET");
+			return (-1);
+		}
+		*pre = '/';
+		if (strlcpy(pre + 1, ruleset.name, len) < len)
+			read_rulesets(dev, ruleset.path);
+		*pre = '\0';
 	}
 
 	return (0);
 }
 
+void
+compute_anchor_field(void)
+{
+	struct anchor_name *a;
+	int sum, cnt, mx, nx;
+	sum = cnt = mx = 0;
+
+	for (a = anchor_root; a != NULL; a = a->next, cnt++) {
+		int len;
+		if (a->ref == 0)
+			continue;
+		len = strlen(a->name);
+		sum += len;
+		if (len > mx)
+			mx = len;
+	}
+
+	nx = sum/cnt;
+	if (nx < ANCHOR_FLD_SIZE)
+		nx = (mx < ANCHOR_FLD_SIZE) ? mx : ANCHOR_FLD_SIZE;
+
+	if (fields[FLD_ANCHOR].max_width != mx ||
+	    fields[FLD_ANCHOR].norm_width != nx) {
+		fields[FLD_ANCHOR].max_width = mx;
+		fields[FLD_ANCHOR].norm_width = nx;
+		field_setup();
+		need_update = 1;
+	}
+}
+#endif
+
+int
+read_rules(int dev)
+{
+	int ret;
+	num_rules = 0;
+
+#ifdef HAVE_RULE_LABELS
+	label_length = MIN_LABEL_SIZE;
+#endif
+
+#ifdef HAVE_RULESETS
+	reset_anchor_names();
+	ret = read_rulesets(dev, NULL);
+	compute_anchor_field();
+#else
+	ret = read_anchor_rules(dev, NULL);
+#endif
+
+#ifdef HAVE_RULE_LABELS
+	{
+		int nw, mw;
+		nw = mw = label_length;
+		if (nw > 16)
+			nw = 16;
+
+		if (fields[FLD_LABEL].norm_width != nw || 
+		    fields[FLD_LABEL].max_width != mw) {
+			fields[FLD_LABEL].norm_width = nw;
+			fields[FLD_LABEL].max_width = mw;
+			fprintf(stderr, "update\n");
+			field_setup();
+			need_update = 1;
+		}
+	}
+#endif
+
+	num_disp = num_rules;
+	return (ret);
+}
+
 #ifdef HAVE_ADDR_WRAP
 void
 tb_print_addrw(struct pf_addr_wrap *addr, struct pf_addr *mask, u_int8_t af)
@@ -1751,6 +1972,10 @@ tb_print_op(u_int8_t op, const char *a1, const char *a2)
 		tbprintf("%s >< %s ", a1, a2);
 	else if (op == PF_OP_XRG)
 		tbprintf("%s <> %s ", a1, a2);
+#ifdef HAVE_OP_RRG
+	else if (op == PF_OP_RRG)
+		tbprintf("%s:%s ", a1, a2);
+#endif
 	else if (op == PF_OP_EQ)
 		tbprintf("= %s ", a1);
 	else if (op == PF_OP_NE)
@@ -1802,7 +2027,11 @@ tb_print_fromto(struct pf_rule_addr *src, struct pf_rule_addr *dst,
 			 PF_AZERO(PT_MASK(src), AF_INET6))
 			tbprintf("any ");
 		else {
+#ifdef HAVE_NEG
+			if (src->neg)
+#else
 			if (src->not)
+#endif
 				tbprintf("! ");
 #ifdef HAVE_ADDR_WRAP
 			tb_print_addrw(&src->addr, PT_MASK(src), af);
@@ -1823,7 +2052,11 @@ tb_print_fromto(struct pf_rule_addr *src, struct pf_rule_addr *dst,
 			 PF_AZERO(PT_MASK(dst), AF_INET6))
 			tbprintf("any ");
 		else {
+#ifdef HAVE_NEG
+			if (dst->neg)
+#else
 			if (dst->not)
+#endif
 				tbprintf("! ");
 #ifdef HAVE_ADDR_WRAP
 			tb_print_addrw(&dst->addr, PT_MASK(dst), af);
@@ -1890,8 +2123,14 @@ print_rule(struct pf_rule *pr)
 #ifdef HAVE_RULE_STATES
 	print_fld_size(FLD_STATS, pr->states);
 #endif
+
+#ifdef HAVE_INOUT_COUNT_RULES
+	print_fld_size(FLD_PKTS, pr->packets[0] + pr->packets[1]);
+	print_fld_size(FLD_BYTES, pr->bytes[0] + pr->bytes[1]);
+#else
 	print_fld_size(FLD_PKTS, pr->packets);
 	print_fld_size(FLD_BYTES, pr->bytes);
+#endif
 	print_fld_uint(FLD_RULE, pr->nr);
 	print_fld_str(FLD_DIR, pr->direction == PF_OUT ? "Out" : "In");
 	if (pr->quick)
@@ -1940,15 +2179,9 @@ print_rule(struct pf_rule *pr)
 
 	tb_start();
 
-#ifdef HAVE_RULE_ANCHOR
-	if (pr->anchorname[0])
-		tbprintf("anchor %s ", pr->anchorname);
-	else {
 #ifdef HAVE_RULE_NATPASS
-		if (pr->natpass)
-			tbprintf("pass ");
-#endif
-	}
+	if (pr->natpass)
+		tbprintf("pass ");
 #endif
 	if (pr->action == PF_DROP) {
 		if (pr->rule_flag & PFRULE_RETURNRST)
@@ -1996,9 +2229,11 @@ print_rule(struct pf_rule *pr)
 		tb_print_flags(pr->flagset);
 	}
 
+	tbprintf(" ");
+
 #ifdef HAVE_RULE_TOS
 	if (pr->tos)
-		tbprintf(" tos 0x%2.2x", pr->tos);
+		tbprintf("tos 0x%2.2x ", pr->tos);
 #endif
 #ifdef PFRULE_FRAGMENT
 	if (pr->rule_flag & PFRULE_FRAGMENT)
@@ -2013,13 +2248,13 @@ print_rule(struct pf_rule *pr)
 		tbprintf("random-id ");
 #endif
 	if (pr->min_ttl)
-		tbprintf(" min-ttl %d", pr->min_ttl);
+		tbprintf("min-ttl %d ", pr->min_ttl);
 #ifdef HAVE_MAX_MSS
 	if (pr->max_mss)
-		tbprintf(" max-mss %d", pr->max_mss);
+		tbprintf("max-mss %d ", pr->max_mss);
 #endif
 	if (pr->allow_opts)
-		tbprintf(" allow-opts");
+		tbprintf("allow-opts ");
 
 	if (pr->action == PF_SCRUB) {
 #ifdef PFRULE_REASSEMBLE_TCP
@@ -2056,6 +2291,10 @@ print_rule(struct pf_rule *pr)
 #endif
 	print_fld_str(FLD_RINFO, tmp_buf);
 
+#ifdef HAVE_RULESETS
+	/* XXX anchor field overloaded with anchor name */
+	print_fld_str(FLD_ANCHOR, (char *)pr->anchor);
+#endif
 	tb_end();
 
 	end_line();
@@ -2282,33 +2521,6 @@ read_queues(int dev)
 	return(0);
 }
 
-#define	R2S_BUFS	8
-#define	RATESTR_MAX	16
-
-char *
-rate2str(double rate)
-{
-	char		*buf;
-	static char	 r2sbuf[R2S_BUFS][RATESTR_MAX];  /* ring bufer */
-	static int	 idx = 0;
-	int		 i;
-	static const char unit[] = " KMG";
-
-	buf = r2sbuf[idx++];
-	if (idx == R2S_BUFS)
-		idx = 0;
-
-	for (i = 0; rate >= 1000 && i <= 3; i++)
-		rate /= 1000;
-
-	if (i == 0)
-		snprintf(buf, RATESTR_MAX, "%d", (int)rate);
-	else
-		snprintf(buf, RATESTR_MAX, "%d%c", (int)rate, unit[i]);
-
-	return (buf);
-}
-
 double
 calc_interval(struct timeval *cur_time, struct timeval *last_time)
 {
@@ -2357,8 +2569,7 @@ print_queue(struct pf_altq_node *node)
 	if (node->altq.scheduler == ALTQT_CBQ ||
 	    node->altq.scheduler == ALTQT_HFSC
 		)
-		print_fld_str(FLD_BANDW,
-			      rate2str((double)node->altq.bandwidth));
+		print_fld_bw(FLD_BANDW, (double)node->altq.bandwidth);
 	
 	if (node->altq.priority != DEFAULT_PRIORITY)
 		print_fld_uint(FLD_PRIO,
@@ -2411,10 +2622,26 @@ print_queue(struct pf_altq_node *node)
 		break;
 	case ALTQT_HFSC:
 		print_fld_str(FLD_SCHED, "hfsc");
+		print_fld_size(FLD_PKTS,
+				node->qstats.data.hfsc_stats.xmit_cnt.packets);
+		print_fld_size(FLD_BYTES,
+				node->qstats.data.hfsc_stats.xmit_cnt.bytes);
+		print_fld_size(FLD_DROPP,
+				node->qstats.data.hfsc_stats.drop_cnt.packets);
+		print_fld_size(FLD_DROPB,
+				node->qstats.data.hfsc_stats.drop_cnt.bytes);
+		print_fld_size(FLD_QLEN, node->qstats.data.hfsc_stats.qlength);
+		if (interval > 0) {
+			pps = calc_pps(node->qstats.data.hfsc_stats.xmit_cnt.packets,
+					node->qstats_last.data.hfsc_stats.xmit_cnt.packets, interval);
+			bps = calc_rate(node->qstats.data.hfsc_stats.xmit_cnt.bytes,
+					node->qstats_last.data.hfsc_stats.xmit_cnt.bytes, interval);
+		}
 		break;
 	}
 
-	if (node->altq.scheduler != ALTQT_HFSC && interval > 0) {
+	/* if (node->altq.scheduler != ALTQT_HFSC && interval > 0) { */
+	if (node->altq.scheduler && interval > 0) {
 		tb_start();
 		if (pps > 0 && pps < 1)
 			tbprintf("%-3.1lf", pps);
@@ -2423,7 +2650,7 @@ print_queue(struct pf_altq_node *node)
 		
 		print_fld_str(FLD_PKTSPS, tmp_buf);
 		tb_end();
-		print_fld_str(FLD_BYTESPS, rate2str(bps));
+		print_fld_bw(FLD_BYTESPS, bps);
 	}
 }
 
@@ -2802,7 +3029,9 @@ main(int argc, char *argv[])
 	int dev = -1;
 	int ch;
 
-	while ((ch = getopt(argc, argv, "abhirc:s:d:o:v:w:")) != -1) {
+	int start = 0;
+
+	while ((ch = getopt(argc, argv, "abhirc:s:d:o:v:w:S:")) != -1) {
 		switch (ch) {
 		case 'a':
 			maxstates = -1;
@@ -2830,6 +3059,11 @@ main(int argc, char *argv[])
 		case 'r':
 			sortdir *= -1;
 			break;
+		case 'S':
+			start = atoi(optarg);
+			if (start < 0)
+				start = 0;
+			break;
 		case 's':
 			delay = atoi(optarg);
 			if (delay < 1)
@@ -2884,7 +3118,7 @@ main(int argc, char *argv[])
 
 	/* preallocate existing states if possible */
 	if (ioctl(dev, DIOCGETSTATUS, &status)) {
-		warnx("DIOCGETSTATUS");
+		warn("DIOCGETSTATUS");
 		alloc_buf(0);
 	} else
 		alloc_buf(status.states);
@@ -2899,7 +3133,7 @@ main(int argc, char *argv[])
 #ifdef HAVE_MAX_STATES
 	show_field(FLD_STMAX);
 #endif
-#ifdef HAVE_RULE_ANCHOR
+#ifdef HAVE_RULESETS
 	show_field(FLD_ANCHOR);
 #endif
 
@@ -2910,6 +3144,8 @@ main(int argc, char *argv[])
 	if (rawmode && countmax == 0)
 		countmax = 1;
 
+	dispstart = start;
+
 	gotsig_alarm = 1;
 	for (;;) {
 		if (gotsig_alarm) {
